"""
A ChoiceEnum to represent ScanningSequence_ values.

.. _ScanningSequence: https://dicom.innolitics.com/ciods/mr-image/mr-image/00180020

"""


from django_dicom.models.choice_enum import ChoiceEnum


class ScanningSequence(ChoiceEnum):
    SE = "Spin Echo"
    IR = "Inversion Recovery"
    GR = "Gradient Recalled"
    EP = "Echo Planar"
    RM = "Research Mode"



import nibabel as nib
import numpy as np
import os

from django.db import models
from django_extensions.db.models import TimeStampedModel


class NIfTI(TimeStampedModel):
    """
    A model for maintaining NIfTI_ files.

    .. _NIfTI: https://nifti.nimh.nih.gov/nifti-1/
    """

    path = models.FilePathField(max_length=500, unique=True)

    # is_raw is meant to be set according to whether the created instance
    # is the product of a direct conversion from the raw data to NIfTI (True),
    # or of a manipulation of the data (False).
    is_raw = models.BooleanField(default=False)

    # As long as this instance is the product of some conversion or manipulation
    # of a single Scan instance, this field is meant to keep a reference to that
    # instance. If it is the product of multiple Scan instances, this field may
    # be set to None, and any associated Scan instances should be reachable
    # through whichever analysis run instance they are represnted with.
    parent = models.ForeignKey(
        "django_mri.Scan",
        on_delete=models.CASCADE,
        related_name="derived_niftis",
        blank=True,
        null=True,
    )

    def get_data(self) -> np.ndarray:
        """
        Uses nibabel_ to return the underlying pixel data as a NumPy_ array.
        
        .. _nibabel: https://nipy.org/nibabel/
        .. _NumPy: http://www.numpy.org/

        Returns
        -------
        np.ndarray
            Pixel data.
        """

        return nib.load(self.path).get_data()

    def get_b_value(self) -> list:
        """
        Returns the degree of diffusion weighting applied (b-value_) for each
        diffusion direction. This method relies on dcm2niix_'s default
        configuration in which when diffusion-weighted images (DWI_) are
        converted, another file with the same name and a "bval" extension is
        created alongside.

        .. _b-value: https://radiopaedia.org/articles/b-values-1
        .. _dcm2niix: https://github.com/rordenlab/dcm2niix
        .. _DWI: https://en.wikipedia.org/wiki/Diffusion_MRI
        
        Returns
        -------
        list
            b-value for each diffusion direction.
        """

        file_name = self.path.replace("nii.gz", "bval")
        if os.path.isfile(file_name):
            with open(file_name, "r") as file_object:
                content = file_object.read()
            content = content.splitlines()[0].split("\t")
            return [int(value) for value in content]
        return None

    def get_b_vector(self) -> list:
        """
        Returns the b-vectors_ representing the diffusion weighting gradient scheme.
        This method relies on dcm2niix_'s default configuration in which when
        diffusion-weighted images (DWI_) are converted, another file with the same
        name and a "bvec" extension is created alongside.

        .. _b-vectors: https://mrtrix.readthedocs.io/en/latest/concepts/dw_scheme.html
        .. _dcm2niix: https://github.com/rordenlab/dcm2niix
        .. _DWI: https://en.wikipedia.org/wiki/Diffusion_MRI
        
        Returns
        -------
        list
            b-value for each diffusion direction
        """
        file_name = self.path.replace("nii.gz", "bvec")
        if os.path.isfile(file_name):
            with open(file_name, "r") as file_object:
                content = file_object.read()
            return [
                [float(value) for value in vector.rstrip().split("\t")]
                for vector in content.rstrip().split("\n")
            ]
        return None

    @property
    def b_value(self) -> list:
        return self.get_b_value()

    @property
    def b_vector(self) -> list:
        return self.get_b_vector()

    @property
    def subject_id(self) -> int:
        """
        If this instance has a single (Scan) origin, returns the related subject's primary key.
        
        Returns
        -------
        int
            The subject ID to which this NIfTI file's origin is related with.
        """

        return self.parent.subject if self.parent else None



import nipype

from django_mri.analysis.specifications.fsl.bet import (
    BET_INPUT_SPECIFICATION,
    BET_OUTPUT_SPECIFICATION,
)
from django_mri.analysis.specifications.fsl.flirt import (
    FLIRT_INPUT_SPECIFICATION,
    FLIRT_OUTPUT_SPECIFICATION,
)
from django_mri.analysis.specifications.fsl.fnirt import (
    FNIRT_INPUT_SPECIFICATION,
    FNIRT_OUTPUT_SPECIFICATION,
)
from nipype.interfaces.fsl import BET, FLIRT, FNIRT

analysis_definitions = [
    {
        "title": "BET",
        "description": "FSL brain extraction (BET).",
        "versions": [
            {
                "title": BET().version,
                "description": f"Default BET version for nipype {nipype.__version__}.",
                "input": BET_INPUT_SPECIFICATION,
                "output": BET_OUTPUT_SPECIFICATION,
                "nested_results_attribute": "outputs.get_traitsfree",
            }
        ],
    },
    {
        "title": "FLIRT",
        "description": "FLIRT (FMRIB's Linear Image Registration Tool) is a fully automated robust and accurate tool for linear (affine) intra- and inter-modal brain image registration.",
        "versions": [
            {
                "title": FLIRT().version,
                "description": f"Default FLIRT version for nipype {nipype.__version__}.",
                "input": FLIRT_INPUT_SPECIFICATION,
                "output": FLIRT_OUTPUT_SPECIFICATION,
                "nested_results_attribute": "outputs.get_traitsfree",
            }
        ],
    },
    {
        "title": "FNIRT",
        "description": "FSL FNIRT wrapper for non-linear registration.",
        "versions": [
            {
                "title": FNIRT().version,
                "description": f"Default FNIRT version for nipype {nipype.__version__}.",
                "input": FNIRT_INPUT_SPECIFICATION,
                "output": FNIRT_OUTPUT_SPECIFICATION,
                "nested_results_attribute": "outputs.get_traitsfree",
            }
        ],
    },
]



# import factory
# import pytz

# from django.db.models.signals import pre_save
# from django_mri.models.nifti import NIfTI
# from django_mri.models.sequence_type import SequenceType
# from django_mri.models.scan import Scan
# from django_mri.models.choices import ScanningSequence, SequenceVariant

# SCANNING_SEQUENCES = [sequence.name for sequence in ScanningSequence]
# SEQUENCE_VARIANTS = [variant.name for variant in SequenceVariant]


# class SequenceTypeFactory(factory.django.DjangoModelFactory):
#     title = factory.Faker("sentence", nb_words=2, variable_nb_words=True)
#     description = factory.Faker("text", max_nb_chars=50)
#     scanning_sequence = factory.Faker(
#         "words", nb=3, ext_word_list=SCANNING_SEQUENCES, unique=True
#     )
#     sequence_variant = factory.Faker("words", nb=2, ext_word_list=SEQUENCE_VARIANTS)

#     class Meta:
#         model = SequenceType


# @factory.django.mute_signals(pre_save)
# class ScanFactory(factory.django.DjangoModelFactory):
#     time = factory.Faker("date_time_this_century", before_now=True, tzinfo=pytz.UTC)
#     description = factory.Faker("sentence", nb_words=5)
#     number = factory.Faker("random_int", min=0, max=300)
#     echo_time = factory.Faker("pyfloat", positive=True, left_digits=2, right_digits=2)
#     repetition_time = factory.Faker("pyfloat", positive=True, left_digits=2)
#     inversion_time = factory.Faker("pyfloat", positive=True, left_digits=2)
#     # spatial_resolution = factory.Faker("pylist", float, nb_elements=3)
#     sequence_type = factory.SubFactory(SequenceTypeFactory)
#     comments = factory.Faker("sentence", nb_words=10)
#     # subject = factory.SubFactory(SubjectFactory)
#     subject = factory.Faker("random_int", min=1, max=5000)
#     _nifti = factory.SubFactory("mri.tests.factories.NIfTIFactory", parent=None)

#     class Meta:
#         model = Scan


# class NIfTIFactory(factory.django.DjangoModelFactory):
#     path = factory.Faker("file_path", depth=4, extension="nii.gz")
#     is_raw = factory.Faker("boolean")
#     parent = factory.SubFactory(ScanFactory)

#     class Meta:
#         model = NIfTI



"""
A list of common MRI sequences and their respective DICOM header attributes.

"""


sequences = [
    {
        "title": "MPRAGE",
        "description": "A T1-weighted 3D magnetization prepared-rapid acquisition gradient echo sequence. See https://pubs.rsna.org/doi/10.1148/radiology.182.3.1535892 for more information.",
        "scanning_sequence": ["GR", "IR"],
        "sequence_variant": ["SK", "SP", "MP"],
    },
    {
        "title": "Localizer",
        "description": "A set of three-plane, low-resolution, large field-of view scans that are obtained to be used for configuring the execution of subsequent scans.",
        "scanning_sequence": ["GR"],
        "sequence_variant": ["SP", "OSP"],
    },
    {
        "title": "IR-EPI",
        "description": "180° inversion pulse to prepare magnetization and then a RF excitation pulse which results in T1-weighting.",
        "scanning_sequence": ["EP", "IR"],
        "sequence_variant": ["SK", "SP", "MP", "OSP"],
    },
    {
        "title": "FLAIR",
        "description": "Fluid-attenuated inversion recovery is an MRI sequence with an inversion recovery set to null fluids. ",
        "scanning_sequence": ["SE", "IR"],
        "sequence_variant": ["SK", "SP", "MP"],
    },
    {
        "title": "DWI",
        "description": "Diffusion-weighted magnetic resonance imaging (DWI or DW-MRI) is the use of specific MRI sequences as well as software that generates images from the resulting data, that uses the diffusion of water molecules to generate contrast in MR images.",
        "scanning_sequence": ["EP"],
        "sequence_variant": ["SK", "SP"],
    },
    {
        "title": "fMRI",
        "description": "Functional magnetic resonance imaging or functional MRI (fMRI) measures brain activity by detecting changes associated with blood flow. This technique relies on the fact that cerebral blood flow and neuronal activation are coupled. When an area of the brain is in use, blood flow to that region also increases.",
        "scanning_sequence": ["EP"],
        "sequence_variant": ["SK", "SS"],
    },
]



import os
import subprocess

BASE_DIR = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
DCM2NIIX = os.path.join(BASE_DIR, "utils", "dcm2niix")


class Dcm2niix:
    FLAGS = {"compressed": "-z", "BIDS": "-b", "name": "-f", "directory": "-o"}
    BOOLEAN = {True: "y", False: "n"}

    def __init__(self, path: str = DCM2NIIX):
        self.path = path

    def generate_command(
        self,
        path: str,
        destination: str,
        compressed: bool = True,
        generate_json: bool = True,
    ) -> list:
        directory, name = os.path.dirname(destination), os.path.basename(destination)
        return [
            self.path,
            self.FLAGS["compressed"],
            self.BOOLEAN[compressed],
            self.FLAGS["BIDS"],
            self.BOOLEAN[generate_json],
            self.FLAGS["directory"],
            directory,
            self.FLAGS["name"],
            name,
            path,
        ]

    def convert(
        self,
        path: str,
        destination: str,
        compressed: bool = True,
        generate_json: bool = True,
    ) -> str:
        command = self.generate_command(
            path, destination, compressed=compressed, generate_json=generate_json
        )
        try:
            subprocess.check_output(command)
            if os.path.isfile(f"{destination}.nii.gz"):
                return f"{destination}.nii.gz"
            else:
                raise RuntimeError(
                    "Failed to create NIfTI file using dcm2niix! Please check application configuration"
                )
        except FileNotFoundError:
            raise NotImplementedError(
                "Could not call dcm2niix! Please check settings configuration."
            )



"""
Choice ENUMs for easier maintenance of CharField's choice parameters. 

"""


from django_mri.models.choices.scanning_sequence import ScanningSequence
from django_mri.models.choices.sequence_variant import SequenceVariant



from django_mri.models import Scan, NIfTI
from rest_framework import serializers


class NiftiSerializer(serializers.HyperlinkedModelSerializer):
    url = serializers.HyperlinkedIdentityField(view_name="mri:nifti-detail")
    parent = serializers.HyperlinkedRelatedField(
        view_name="mri:scan-detail", queryset=Scan.objects.all()
    )

    class Meta:
        model = NIfTI
        exclude = ("path",)



import os

from django.test import TestCase
from django_dicom.models import Series
from django_mri.data_import.local_import import LocalImport
from django_mri.models import NIfTI
from tests.fixtures import TEST_FILES_PATH


class LocalImportTestCase(TestCase):
    def setUp(self):
        self.importer = LocalImport(TEST_FILES_PATH)

    def test_initialization(self):
        result = self.importer.path
        expected = TEST_FILES_PATH
        self.assertEqual(result, expected)

    def test_path_generator_without_extension(self):
        """
        Tests the :meth:`~django_mri.data_import.local_import.LocalImport.path_generator`
        method with no *extension* parameter setting.
        
        """

        counter = 0
        for path in LocalImport(TEST_FILES_PATH).path_generator():
            is_valid_path = os.path.isfile(path)
            self.assertTrue(is_valid_path)
            is_under_base_dir = path.startswith(TEST_FILES_PATH)
            self.assertTrue(is_under_base_dir)
            counter += 1
        self.assertEqual(counter, 40)

    def test_path_generator_with_extension(self):
        """
        Tests the :meth:`~django_mri.data_import.local_import.LocalImport.path_generator`
        method with the *extension* parameter set.
        
        """

        # A dictionary of extensions and the number of files we expect
        extensions = {"nii.gz": 1, "dcm": 39}

        for extension in extensions:
            counter = 0
            generator = LocalImport(TEST_FILES_PATH).path_generator(extension=extension)
            for path in generator:
                is_valid_path = os.path.isfile(path)
                self.assertTrue(is_valid_path)
                is_under_base_dir = path.startswith(TEST_FILES_PATH)
                self.assertTrue(is_under_base_dir)
                counter += 1
            self.assertEqual(counter, extensions.get(extension))

    def test_import_nifti_files(self):
        self.importer.import_nifti_files(verbose=False)
        self.assertEqual(NIfTI.objects.count(), 1)
        NIfTI.objects.all().delete()

    def test_run(self):
        self.importer.run(verbose=False)
        self.assertEqual(NIfTI.objects.count(), 1)
        self.assertEqual(Series.objects.count(), 2)



from django.test import TestCase
from django_mri.models.choices import ScanningSequence, SequenceVariant
from django_mri.models.sequence_type import SequenceType


TEST_SEQUENCE = {
    "title": "TEST",
    "scanning_sequence": [ScanningSequence.GR.name, ScanningSequence.RM.name],
    "sequence_variant": [
        SequenceVariant.SK.name,
        SequenceVariant.SP.name,
        SequenceVariant.MP.name,
    ],
}


class SequenceTypeModelTestCase(TestCase):
    @classmethod
    def setUpTestData(cls):
        SequenceType.objects.create(**TEST_SEQUENCE)

    def setUp(self):
        self.sequence_type = SequenceType.objects.first()

    ########
    # Meta #
    ########

    def test_unique_together(self):
        result = SequenceType._meta.unique_together
        expected = (("scanning_sequence", "sequence_variant"),)
        self.assertTupleEqual(result, expected)

    ##########
    # Fields #
    ##########

    # title
    def test_title_blank_and_null(self):
        field = SequenceType._meta.get_field("title")
        self.assertFalse(field.blank)
        self.assertFalse(field.null)

    def test_title_uniqueness(self):
        field = SequenceType._meta.get_field("title")
        self.assertFalse(field.unique)

    # description
    def test_description_blank_and_null(self):
        field = SequenceType._meta.get_field("description")
        self.assertTrue(field.blank)
        self.assertTrue(field.null)

    def test_description_uniqueness(self):
        field = SequenceType._meta.get_field("description")
        self.assertFalse(field.unique)

    # scanning_sequence
    def test_scanning_sequence_blank_and_null(self):
        field = SequenceType._meta.get_field("scanning_sequence")
        self.assertTrue(field.blank)
        self.assertTrue(field.null)

    def test_scanning_sequence_size(self):
        field = SequenceType._meta.get_field("scanning_sequence")
        self.assertEqual(field.size, 5)

    def test_scanning_sequence_base_max_length(self):
        field = SequenceType._meta.get_field("scanning_sequence")
        self.assertEqual(field.base_field.max_length, 2)

    def test_scanning_sequence_choices(self):
        field = SequenceType._meta.get_field("scanning_sequence")
        result = field.base_field.choices
        expected = ScanningSequence.choices()
        self.assertTupleEqual(result, expected)

    # sequence_variant
    def test_sequence_variant_blank_and_null(self):
        field = SequenceType._meta.get_field("sequence_variant")
        self.assertTrue(field.blank)
        self.assertTrue(field.null)

    def test_sequence_variant_base_max_length(self):
        field = SequenceType._meta.get_field("sequence_variant")
        self.assertEqual(field.base_field.max_length, 4)

    def test_sequence_variant_choices(self):
        field = SequenceType._meta.get_field("sequence_variant")
        result = field.base_field.choices
        expected = SequenceVariant.choices()
        self.assertTupleEqual(result, expected)

    ###########
    # Methods #
    ###########

    def test_get_string(self):
        result = str(self.sequence_type)
        expected = TEST_SEQUENCE["title"]
        self.assertEqual(result, expected)

    ####################
    # Common Sequences #
    ####################

    def test_common_sequences_load(self):
        from django_mri.models.common_sequences import sequences

        objects = [SequenceType(**sequence) for sequence in sequences]
        SequenceType.objects.bulk_create(objects)



from django_mri.models.sequence_type import SequenceType
from rest_framework import serializers


class SequenceTypeSerializer(serializers.HyperlinkedModelSerializer):
    url = serializers.HyperlinkedIdentityField(view_name="mri:sequencetype-detail")

    class Meta:
        model = SequenceType
        fields = (
            "id",
            "title",
            "description",
            "scanning_sequence",
            "sequence_variant",
            "url",
        )



from django_mri.filters.sequence_type_filter import SequenceTypeFilter
from django_mri.models.sequence_type import SequenceType
from django_mri.serializers import SequenceTypeSerializer
from django_mri.views.defaults import DefaultsMixin
from django_mri.views.pagination import StandardResultsSetPagination
from rest_framework import viewsets


class SequenceTypeViewSet(DefaultsMixin, viewsets.ModelViewSet):
    pagination_class = StandardResultsSetPagination
    queryset = SequenceType.objects.all()
    serializer_class = SequenceTypeSerializer
    filter_class = SequenceTypeFilter



from django.apps import apps
from django.conf import settings
from django.contrib.auth import get_user_model
from django_dicom.data_import.import_image import ImportImage
from django_mri.models.scan import Scan
from io import BufferedReader


subject_app_label, subject_model_name = settings.SUBJECT_MODEL.split(".")
Subject = apps.get_model(app_label=subject_app_label, model_name=subject_model_name)
User = get_user_model()


class ImportScan:
    def __init__(
        self,
        subject: Subject,
        scan_file: BufferedReader,
        scan_type: str = "dcm",
        user: User = None,
    ):
        self.subject = subject
        self.scan_file = scan_file
        self.scan_type = scan_type
        self.user = user

    def import_dicom_image(self):
        dicom_image, _ = ImportImage(self.scan_file).run()
        scan, created = Scan.objects.get_or_create(dicom=dicom_image.series)
        if created:
            scan.update_fields_from_dicom()
            if self.user:
                scan.added_by = self.user
        scan.subject = self.subject
        scan.save()
        return scan, created

    def run(self) -> tuple:
        if self.scan_type == "dcm":
            return self.import_dicom_image()
        else:
            raise ValueError("Invalid scan type!")




import os
import pytz

from datetime import datetime
from django.contrib.auth import get_user_model
from django.conf import settings
from django.contrib.postgres.fields import ArrayField
from django.core.validators import MinValueValidator
from django.db import models
from django_extensions.db.models import TimeStampedModel
from django_mri.interfaces.dcm2niix import Dcm2niix
from django_mri.models.nifti import NIfTI
from django_mri.models.sequence_type import SequenceType


class Scan(TimeStampedModel):
    """
    A model used to represent an MRI scan independently from the file-format in
    which it is saved. This model handles any conversions between formats in case
    they are required, and allows for easy querying of MRI scans based on universal
    attributes.

    """

    institution_name = models.CharField(max_length=64, blank=True, null=True)
    time = models.DateTimeField(
        blank=True, null=True, help_text="The time in which the scan was acquired."
    )
    description = models.CharField(
        max_length=100,
        blank=True,
        null=True,
        help_text="A short description of the scan's acqusition parameters.",
    )
    number = models.IntegerField(
        blank=True,
        null=True,
        validators=[MinValueValidator(0)],
        help_text="The number of this scan relative to the session in which it was acquired.",
    )

    # Relatively universal MRI scan attributes. These might be infered from the
    # raw file's meta-data.
    echo_time = models.FloatField(
        blank=True,
        null=True,
        validators=[MinValueValidator(0)],
        help_text="The time between the application of the radiofrequency excitation pulse and the peak of the signal induced in the coil (in milliseconds).",
    )
    repetition_time = models.FloatField(
        blank=True,
        null=True,
        validators=[MinValueValidator(0)],
        help_text="The time between two successive RF pulses (in milliseconds).",
    )
    inversion_time = models.FloatField(
        blank=True,
        null=True,
        validators=[MinValueValidator(0)],
        help_text="The time between the 180-degree inversion pulse and the following spin-echo (SE) sequence (in milliseconds).",
    )
    spatial_resolution = ArrayField(models.FloatField(), size=3, blank=True, null=True)

    comments = models.TextField(
        max_length=1000,
        blank=True,
        null=True,
        help_text="If anything noteworthy happened during acquisition, it may be noted here.",
    )

    # If this instance's origin is a DICOM file, or it was saved as one, this field
    # keeps the relation to that django_dicom.Series instance.
    dicom = models.OneToOneField(
        "django_dicom.Series",
        on_delete=models.PROTECT,
        blank=True,
        null=True,
        related_name="scan",
        verbose_name="DICOM Series",
    )
    # Keep track of whether we've updated the instance's fields from the DICOM
    # header data.
    is_updated_from_dicom = models.BooleanField(default=False)

    # If converted to NIfTI, keep a reference to the resulting instance.
    # The reason it is suffixed with an underline is to allow for "nifti"
    # to be used as a property that automatically returns an existing instance
    # or creates one.
    nifti = models.OneToOneField(
        "django_mri.NIfTI", on_delete=models.SET_NULL, blank=True, null=True
    )

    subject = models.ForeignKey(
        settings.SUBJECT_MODEL,
        on_delete=models.PROTECT,
        related_name="mri_scans",
        blank=True,
        null=True,
    )

    study_groups = models.ManyToManyField(
        settings.STUDY_GROUP_MODEL, related_name="mri_scans", blank=True
    )

    added_by = models.ForeignKey(
        get_user_model(),
        related_name="mri_uploads",
        blank=True,
        null=True,
        on_delete=models.SET_NULL,
    )

    class Meta:
        ordering = ("time",)
        verbose_name_plural = "MRI Scans"

    def update_fields_from_dicom(self) -> None:
        """
        Sets instance fields from related DICOM series.
        TODO: Needs refactoring.

        Raises
        ------
        AttributeError
            If not DICOM series is related to this scan.
        """

        if self.dicom:
            self.institution_name = self.dicom.institution_name
            self.number = self.dicom.number
            self.time = datetime.combine(
                self.dicom.date, self.dicom.time, tzinfo=pytz.UTC
            )
            self.description = self.dicom.description
            self.echo_time = self.dicom.echo_time
            self.inversion_time = self.dicom.inversion_time
            self.repetition_time = self.dicom.repetition_time
            self.spatial_resolution = self.get_spatial_resolution_from_dicom()
            self.is_updated_from_dicom = True
        else:
            raise AttributeError(f"No DICOM data associated with MRI scan {self.id}!")

    def get_spatial_resolution_from_dicom(self) -> list:
        """
        Returns the spatial resolution of the MRI scan as infered from a
        related DICOM series. In DICOM headers, "*x*" and "*y*" resolution
        (the rows and columns of each instance) are listed as "`Pixel Spacing`_"
        and the "*z*" plane resolution corresponds to "`Slice Thickness`_".
        `Pixel Spacing`_ is a `required (type 1)`_ DICOM attribute, and therefore
        has to be returned, however `Slice Thickness`_ may be empty (`type 2`).

        .. _Pixel Spacing: https://dicom.innolitics.com/ciods/mr-image/image-plane/00280030
        .. _Slice Thickness: https://dicom.innolitics.com/ciods/mr-image/image-plane/00180050
        .. _required (type 1): http://dicom.nema.org/dicom/2013/output/chtml/part05/sect_7.4.html
        .. _type 2: http://dicom.nema.org/dicom/2013/output/chtml/part05/sect_7.4.html

        Returns
        -------
        list
            "*[x, y, z]*" spatial resolution in millimeters.
        """

        sample_image = self.dicom.image_set.first()
        slice_thickness = sample_image.header.get_value("SliceThickness")
        if slice_thickness:
            return self.dicom.pixel_spacing + [slice_thickness]
        else:
            return self.dicom.pixel_spacing

    def infer_sequence_type_from_dicom(self) -> SequenceType:
        """
        Returns the appropriate :class:`django_mri.SequenceType` instance according to
        the scan's "*ScanningSequence*" and "*SequenceVariant*" header values.


        Returns
        -------
        SequenceType
            A SequenceType instance.
        """

        try:
            return SequenceType.objects.get(
                scanning_sequence=self.dicom.scanning_sequence,
                sequence_variant=self.dicom.sequence_variant,
            )
        except models.ObjectDoesNotExist:
            return None

    def infer_sequence_type(self) -> SequenceType:
        if self.dicom:
            return self.infer_sequence_type_from_dicom()

    def get_default_nifti_dir(self) -> str:
        """
        Returns the default location for the creation of a NIfTI version of the
        scan. Currently only conversion from DICOM is supported.

        Returns
        -------
        str
            Default location for conversion output.
        """

        if self.dicom:
            return self.dicom.get_path().replace("DICOM", "NIfTI")

    def get_default_nifti_name(self) -> str:
        """
        Returns the default file name for a NIfTI version of this scan.

        Returns
        -------
        str
            Default file name.
        """

        return str(self.id)

    def get_default_nifti_destination(self) -> str:
        """
        Returns the default path for a NIfTI version of this scan.

        Returns
        -------
        str
            Default path for NIfTI file.
        """

        directory = self.get_default_nifti_dir()
        name = self.get_default_nifti_name()
        return os.path.join(directory, name)

    def dicom_to_nifti(
        self,
        destination: str = None,
        compressed: bool = True,
        generate_json: bool = False,
    ) -> NIfTI:
        """
        Convert this scan from DICOM to NIfTI using _dcm2niix.

        .. _dcm2niix: https://github.com/rordenlab/dcm2niix

        Parameters
        ----------
        destination : str, optional
            The desired path for conversion output (the default is None, which
            will create the file in some default location).

        Raises
        ------
        AttributeError
            If no DICOM series is related to this scan.

        Returns
        -------
        NIfTI
            A :class:`django_mri.NIfTI` instance referencing the conversion output.
        """

        if self.dicom:
            dcm2niix = Dcm2niix()
            destination = destination or self.get_default_nifti_destination()
            os.makedirs(os.path.dirname(destination), exist_ok=True)
            nifti_path = dcm2niix.convert(
                self.dicom.get_path(),
                destination,
                compressed=compressed,
                generate_json=generate_json,
            )
            nifti = NIfTI.objects.create(path=nifti_path, parent=self, is_raw=True)
            return nifti
        else:
            raise AttributeError(
                f"Failed to convert scan #{self.id} from DICOM to NIfTI! No DICOM series is related to this scan."
            )

    @property
    def sequence_type(self) -> SequenceType:
        return self.infer_sequence_type()



import numpy as np

from django.test import TestCase
from django_dicom.data_import import LocalImport
from django_mri.models.nifti import NIfTI
from django_mri.models.scan import Scan
from tests.fixtures import (
    DICOM_SERIES_PATH,
    SIEMENS_DWI_SERIES,
    SIEMENS_DWI_SERIES_PATH,
)


class NIfTIModelTestCase(TestCase):
    @classmethod
    def setUpTestData(cls):
        LocalImport(DICOM_SERIES_PATH).run(verbose=False)
        LocalImport(SIEMENS_DWI_SERIES_PATH).run(verbose=False)

    def setUp(self):
        self.simple_scan = Scan.objects.first()
        self.dwi_scan = Scan.objects.last()
        if not self.simple_scan or not self.dwi_scan:
            self.fail("Test scan not created! Check signals.")
        self.simple_nifti = self.simple_scan.nifti
        self.dwi_nifti = self.dwi_scan.nifti

    ##########
    # Fields #
    ##########

    # path
    def test_path_blank_and_null(self):
        field = NIfTI._meta.get_field("path")
        self.assertFalse(field.blank)
        self.assertFalse(field.null)

    def test_path_max_length(self):
        field = NIfTI._meta.get_field("path")
        self.assertEqual(field.max_length, 500)

    def test_path_value(self):
        destination = self.dwi_scan.get_default_nifti_destination()
        expected = destination + ".nii.gz"
        self.assertEqual(self.dwi_nifti.path, expected)

    # is_raw
    def test_is_raw_blank_and_null(self):
        field = NIfTI._meta.get_field("is_raw")
        self.assertFalse(field.blank)
        self.assertFalse(field.null)

    def test_is_raw_default(self):
        field = NIfTI._meta.get_field("is_raw")
        self.assertFalse(field.default)

    # parent
    def test_parent_blank_and_null(self):
        field = NIfTI._meta.get_field("parent")
        self.assertTrue(field.blank)
        self.assertTrue(field.null)

    ###########
    # Methods #
    ###########

    def test_get_data(self):
        data = self.dwi_nifti.get_data()
        self.assertIsInstance(data, np.ndarray)

    def test_get_b_value(self):
        result = self.dwi_nifti.get_b_value()
        self.assertListEqual(result, SIEMENS_DWI_SERIES["b_value"])

    def test_get_b_vector(self):
        result = self.dwi_nifti.get_b_vector()
        self.assertListEqual(result, SIEMENS_DWI_SERIES["b_vector"])

    ##############
    # Properties #
    ##############

    def test_b_value(self):
        result = self.dwi_nifti.b_value
        expected = self.dwi_nifti.get_b_value()
        self.assertListEqual(result, expected)

    def test_b_value_for_non_DWI_returns_none(self):
        self.assertIsNone(self.simple_nifti.b_value)

    def test_b_vector(self):
        result = self.dwi_nifti.b_vector
        expected = self.dwi_nifti.get_b_vector()
        self.assertListEqual(result, expected)

    def test_b_vector_for_non_DWI_returns_none(self):
        self.assertIsNone(self.simple_nifti.b_vector)

    def test_subject_id(self):
        subject_id = "TESTSUB"
        self.simple_scan.subject_id = subject_id
        self.assertEqual(self.simple_nifti.subject_id, subject_id)

    def test_subject_id_with_no_subject(self):
        self.assertIsNone(self.dwi_nifti.subject_id)



import os
import zipfile

from django.apps import apps
from django.contrib.auth import get_user_model
from django.conf import settings
from django_mri.data_import.import_scan import ImportScan


subject_app_label, subject_model_name = settings.SUBJECT_MODEL.split(".")
Subject = apps.get_model(app_label=subject_app_label, model_name=subject_model_name)
User = get_user_model()


class LocalImport:
    """
    Import new MRI data to the database. Currently only two file formats are
    supported; DICOM and NIfTI. The result of running this is the creation of new
    DICOM and NIfTI instances, but **NOT** of :class:`~django_mri.models.scan.Scan`
    instances. These are meant to be created after review by a researcher.
    
    """

    def __init__(self, subject: Subject, path: str, user: User = None):
        """
        Initializes the :class:`~django_mri.data_import.local_import.LocalImport`
        class with the base directory path.
        
        Parameters
        ----------
        subject : Subject
            The research subject to whom the scan belongs.
        path : str
            Base directory for local MRI data import.
        """

        self.subject = subject
        self.path = path
        self.user = user

    @classmethod
    def import_local_dcm(cls, subject: Subject, path: str, user: User = None) -> tuple:
        """
        Imports a single DICOM image (.dcm file) using *django_dicom*'s
        :class:`~django_dicom.data_import.import_scan.ImportScan` class.
        
        Parameters
        ----------
        subject : Subject
            The research subject to whom the scan belongs.
        path : str
            Full path of the DICOM image.
        
        Returns
        -------
        tuple
            (Resulting :class:`~django_dicom.models.image.Image` instance, Created [bool])
        """

        with open(path, "rb") as dcm_buffer:
            return ImportScan(subject, dcm_buffer, scan_type="dcm", user=user).run()

    def import_zip_archive(cls, subject: Subject, path: str, user: User = None) -> None:
        """
        Iterates over the files within a ZIP archive and imports any "*.dcm*" files.
        
        Parameters
        ----------
        path : str
            Local ZIP archive path.
        verbose : bool, optional
            Show a progressbar (default to True).
        """

        with zipfile.ZipFile(path, "r") as archive:
            for file_name in archive.namelist():
                if file_name.endswith(".dcm"):
                    with archive.open(file_name) as dcm_buffer:
                        ImportScan(
                            subject, dcm_buffer, scan_type="dcm", user=user
                        ).run()

    def path_generator(self, extension: str = "") -> str:
        """
        Generates paths from the given directory tree.

        Parameters
        ----------
        extension : str
            A file extension to filter the generated files with.
        
        Returns
        -------
        str
            File path.
        """

        for directory, _, files in os.walk(self.path):
            if extension:
                files = [f for f in files if f.endswith(f".{extension}")]
            for file_name in files:
                yield os.path.join(directory, file_name)

    # def import_nifti_files(self, verbose: bool = True):
    #     """
    #     Imports NIfTI_ (*.nii* files) from the given directory tree.

    #     .. _NIfTI: https://nifti.nimh.nih.gov/

    #     Parameters
    #     ----------
    #     verbose : bool, optional
    #         Prints user friendly information during the run, by default True
    #     """

    #     if verbose:
    #         print("\nImporting NIfTI image files:")

    #     # NIfTI files are often compressed
    #     uncompressed_nii_files = self.path_generator(extension="nii")
    #     compressed_nii_files = self.path_generator(extension="nii.gz")
    #     nii_files = itertools.chain(uncompressed_nii_files, compressed_nii_files)

    #     # Create NIfTI instances in the database
    #     for nii in nii_files:
    #         if verbose:
    #             print(f"Importing {nii}...", end="\t")
    #         NIfTI.objects.create(path=nii, is_raw=True)
    #         if verbose:
    #             print("done!")

    def run(self):
        """
        Import all supported MRI data files from the given directory tree.
        
        """

        if os.path.isdir(self.path):
            dcm_generator = self.path_generator(extension="dcm")
            for dcm_path in dcm_generator:
                self.import_local_dcm(self.subject, dcm_path, user=self.user)
            archives = self.path_generator(extension="zip")
            for archive in archives:
                self.import_zip_archive(self.subject, archive, user=self.user)

        elif os.path.isfile(self.path):
            if self.path.endswith(".dcm"):
                self.import_local_dcm(self.subject, self.path, user=self.user)
            elif self.path.endswith(".zip"):
                self.import_zip_archive(self.subject, self.path, user=self.user)



from django_mri.models import NIfTI
from django_mri.serializers import NiftiSerializer
from django_mri.views.defaults import DefaultsMixin
from django_mri.views.pagination import StandardResultsSetPagination
from rest_framework import viewsets


class NiftiViewSet(DefaultsMixin, viewsets.ModelViewSet):
    pagination_class = StandardResultsSetPagination
    queryset = NIfTI.objects.all()
    serializer_class = NiftiSerializer



from django.db import models
from django_extensions.db.models import TitleDescriptionModel
from django_mri.models.choices import ScanningSequence, SequenceVariant
from django_mri.models.fields import ChoiceArrayField


class SequenceType(TitleDescriptionModel):
    """
    A model to represent familiar MRI sequences. The purpose of this model is to
    provide a title and description for commonly used sequences, as well as to
    facilitate queries.
    Each particular sequence is defined as a unique combination of DICOM's
    "*ScanningSequence*" and "*SequenceVariant*" attributes.

    """

    scanning_sequence = ChoiceArrayField(
        models.CharField(max_length=2, choices=ScanningSequence.choices()),
        size=5,
        blank=True,
        null=True,
    )
    sequence_variant = ChoiceArrayField(
        models.CharField(max_length=4, choices=SequenceVariant.choices()),
        blank=True,
        null=True,
    )

    class Meta:
        unique_together = ("scanning_sequence", "sequence_variant")
        ordering = ("title",)

    def __str__(self) -> str:
        return self.title



import environ
import os

env = environ.Env(
    DEBUG=(bool, False),
    ALLOWED_HOSTS=(list, ["*"]),
    SECRET_KEY=(str, "asdf5sag231sd$#%SADF2341a"),
    BOKEH_SECRET_KEY=(str, ""),
    BOKEH_SIGN_SESSIONS=(bool, True),
    DB_NAME=(str, ""),
    DB_USER=(str, ""),
    DB_PASSWORD=(str, ""),
)

BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
SECRET_KEY = env("SECRET_KEY")
INSTALLED_APPS = [
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.staticfiles",
    "django_extensions",
    "django_dicom",
    "django_mri",
    "tests",
]

MIDDLEWARE = [
    "django.middleware.security.SecurityMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
]

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.postgresql",
        "NAME": env("DB_NAME"),
        "USER": env("DB_USER"),
        "PASSWORD": env("DB_PASSWORD"),
        "HOST": "localhost",
        "PORT": "",
    }
}

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [os.path.join(BASE_DIR, "templates")],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
                "django.template.context_processors.media",
            ]
        },
    }
]

LANGUAGE_CODE = "en-GB"

TIME_ZONE = "UTC"

USE_I18N = True

USE_L10N = True

USE_TZ = True

STATIC_URL = "/static/"

STATICFILES_DIRS = [
    os.path.join(BASE_DIR, "static"),
    os.path.join(BASE_DIR, "node_modules"),
]

MEDIA_ROOT = os.path.join(BASE_DIR, "tests")
MEDIA_URL = "/media/"
ROOT_URLCONF = "django_mri.urls"

REST_FRAMEWORK = {
    # Use Django's standard `django.contrib.auth` permissions,
    # or allow read-only access for unauthenticated users.
    "DEFAULT_PERMISSION_CLASSES": [
        "rest_framework.permissions.DjangoModelPermissionsOrAnonReadOnly"
    ],
    "DEFAULT_PAGINATION_CLASS": "rest_framework.pagination.PageNumberPagination",
    "PAGE_SIZE": 20,
}



import os

from django.core.exceptions import ValidationError
from django.test import TestCase
from django_dicom.data_import import LocalImport
from django_mri.models import Scan, NIfTI
from tests.fixtures import SIEMENS_DWI_SERIES, SIEMENS_DWI_SERIES_PATH


class ScanModelTestCase(TestCase):
    @classmethod
    def setUpTestData(cls):
        LocalImport(SIEMENS_DWI_SERIES_PATH).run(verbose=False)

    def setUp(self):
        self.scan = Scan.objects.last()
        if not self.scan:
            self.fail("Test scan not created! Check signals.")

    ########
    # Meta #
    ########

    def test_ordering(self):
        self.assertTupleEqual(Scan._meta.ordering, ("time",))

    def test_verbose_name_plural(self):
        self.assertTrue(Scan._meta.verbose_name_plural, "MRI Scans")

    ##########
    # Fields #
    ##########

    # time
    def test_time_value(self):
        result = self.scan.time
        expected = SIEMENS_DWI_SERIES["time"]
        self.assertEqual(result, expected)

    def test_time_blank_and_null(self):
        field = Scan._meta.get_field("time")
        self.assertTrue(field.blank)
        self.assertTrue(field.null)

    # description
    def test_description_value(self):
        expected = SIEMENS_DWI_SERIES["description"]
        self.assertEqual(self.scan.description, expected)

    def test_description_blank_and_null(self):
        field = Scan._meta.get_field("description")
        self.assertTrue(field.blank)
        self.assertTrue(field.null)

    def test_description_max_length(self):
        field = Scan._meta.get_field("description")
        self.assertEqual(field.max_length, 100)

    # number
    def test_number_value(self):
        expected = SIEMENS_DWI_SERIES["number"]
        self.assertEqual(self.scan.number, expected)

    def test_number_blank_and_null(self):
        field = Scan._meta.get_field("number")
        self.assertTrue(field.blank)
        self.assertTrue(field.null)

    def test_number_must_be_positive(self):
        invalid = [-100, -40, -1]
        for number in invalid:
            self.scan.number = number
            with self.assertRaises(ValidationError):
                self.scan.full_clean()

    # echo_time
    def test_echo_time_value(self):
        expected = SIEMENS_DWI_SERIES["echo_time"]
        self.assertEqual(self.scan.echo_time, expected)

    def test_echo_time_blank_and_null(self):
        field = Scan._meta.get_field("echo_time")
        self.assertTrue(field.blank)
        self.assertTrue(field.null)

    def test_echo_time_must_be_positive(self):
        invalid = [-14.4, -40.8, -1]
        for number in invalid:
            self.scan.echo_time = number
            with self.assertRaises(ValidationError):
                self.scan.full_clean()

    # repetition_time
    def test_repetition_time_value(self):
        expected = SIEMENS_DWI_SERIES["repetition_time"]
        self.assertEqual(self.scan.repetition_time, expected)

    def test_repetition_time_blank_and_null(self):
        field = Scan._meta.get_field("repetition_time")
        self.assertTrue(field.blank)
        self.assertTrue(field.null)

    def test_repetition_time_must_be_positive(self):
        invalid = [-14.4, -40.8, -1]
        for number in invalid:
            self.scan.repetition_time = number
            with self.assertRaises(ValidationError):
                self.scan.full_clean()

    # inversion_time
    def test_inversion_time_value(self):
        self.assertIsNone(self.scan.inversion_time)

    def test_inversion_time_blank_and_null(self):
        field = Scan._meta.get_field("inversion_time")
        self.assertTrue(field.blank)
        self.assertTrue(field.null)

    def test_inversion_time_must_be_positive(self):
        invalid = [-14.4, -40.8, -1]
        for number in invalid:
            self.scan.inversion_time = number
            with self.assertRaises(ValidationError):
                self.scan.full_clean()

    # spatial_resolution
    def test_spatial_resolution_value(self):
        result = self.scan.spatial_resolution
        expected = SIEMENS_DWI_SERIES["spatial_resolution"]
        self.assertEqual(result, expected)

    def test_spatial_resolution_blank_and_null(self):
        field = Scan._meta.get_field("spatial_resolution")
        self.assertTrue(field.blank)
        self.assertTrue(field.null)

    def test_spatial_resolution_size(self):
        field = Scan._meta.get_field("spatial_resolution")
        self.assertEqual(field.size, 3)

    # sequence_type
    def test_sequence_type_blank_and_null(self):
        field = Scan._meta.get_field("sequence_type")
        self.assertTrue(field.blank)
        self.assertTrue(field.null)

    # comments
    def test_comments_blank_and_null(self):
        field = Scan._meta.get_field("comments")
        self.assertTrue(field.blank)
        self.assertTrue(field.null)

    def test_comments_max_length(self):
        field = Scan._meta.get_field("comments")
        self.assertEqual(field.max_length, 1000)

    # dicom
    def test_dicom_blank_and_null(self):
        field = Scan._meta.get_field("dicom")
        self.assertTrue(field.blank)
        self.assertTrue(field.null)

    def test_dicom_verbose_name(self):
        field = Scan._meta.get_field("dicom")
        expected = "DICOM Series"
        self.assertEqual(field.verbose_name, expected)

    # is_updated_from_dicom
    def test_is_updated_from_dicom_default(self):
        field = Scan._meta.get_field("is_updated_from_dicom")
        self.assertFalse(field.default)

    # _nifti
    def test__nifti_blank_and_null(self):
        field = Scan._meta.get_field("_nifti")
        self.assertTrue(field.blank)
        self.assertTrue(field.null)

    # subject_id
    def test_subject_id_blank_and_null(self):
        field = Scan._meta.get_field("subject_id")
        self.assertTrue(field.blank)
        self.assertTrue(field.null)

    ###########
    # Methods #
    ###########

    def test_update_fields_from_dicom_with_no_dicom_raises_attribute_error(self):
        self.scan.dicom = None
        with self.assertRaises(AttributeError):
            self.scan.update_fields_from_dicom()

    def test_get_spatial_resolution_from_dicom(self):
        result = self.scan.spatial_resolution
        expected = SIEMENS_DWI_SERIES["spatial_resolution"]
        self.assertListEqual(result, expected)

    # TODO: Add test for 2D scan with no "SliceThickness"
    # def test_get_spatial_resolution_from_dicom_without_slice_thickness_attribute(self):
    #     result = self.scan.spatial_resolution
    #     expected = TWO_DIM_TEST_SERIES["spatial_resolution"]
    #     self.assertListEqual(result, expected)

    def test_infer_sequence_type_from_dicom_returns_none(self):
        result = self.scan.infer_sequence_type_from_dicom()
        self.assertIsNone(result)

    def test_get_default_nifti_dir(self):
        result = self.scan.get_default_nifti_dir()
        expected = self.scan.dicom.get_path().replace("DICOM", "NIfTI")
        self.assertEqual(result, expected)

    def test_get_default_nifti_dir_without_dicom(self):
        self.scan.dicom = None
        result = self.scan.get_default_nifti_dir()
        self.assertIsNone(result)

    def test_default_nifti_name(self):
        name = self.scan.get_default_nifti_name()
        expected = str(self.scan.id)
        self.assertEqual(name, expected)

    def test_get_default_nifti_destination(self):
        result = self.scan.get_default_nifti_destination()
        directory = self.scan.get_default_nifti_dir()
        name = self.scan.get_default_nifti_name()
        expected = os.path.join(directory, name)
        self.assertEqual(result, expected)

    def test_dicom_to_nifti_with_no_dicom_raises_attribute_error(self):
        self.scan.dicom = None
        with self.assertRaises(AttributeError):
            self.scan.dicom_to_nifti()

    def test_dicom_to_nifti(self):
        nifti = self.scan.dicom_to_nifti()
        self.assertIsInstance(nifti, NIfTI)

    ##############
    # Properties #
    ##############

    def test_nifti_property_returns_nifti_instance(self):
        result = self.scan.nifti
        self.assertIsInstance(result, NIfTI)
        again = self.scan.nifti
        self.assertEqual(result, again)

    def test_nifti_property_with_conversion_from_dicom(self):
        self.assertIsNone(self.scan._nifti)
        self.assertIsInstance(self.scan.nifti, NIfTI)
        self.assertIsInstance(self.scan._nifti, NIfTI)

    def test_nifti_property_with_no_dicom(self):
        self.scan.dicom = None
        self.assertIsNone(self.scan._nifti)
        with self.assertRaises(AttributeError):
            self.scan.nifti




"""
A ChoiceEnum to represent SequenceVariant_ values.

.. _SequenceVariant: https://dicom.innolitics.com/ciods/mr-image/mr-image/00180021

"""

from django_dicom.models.choice_enum import ChoiceEnum


class SequenceVariant(ChoiceEnum):
    SK = "Segmented k-Space"
    MTC = "Magnetization Transfer Contrast"
    SS = "Steady State"
    TRSS = "Time Reversed Steady State"
    SP = "Spoiled"
    MP = "MAG Prepared"
    OSP = "Oversampling Phase"
    NONE = "None"



"""
An app to manage MRI data. A Scan instance may be related to multiple file
formats and is meant to provide a unified model for dealing with the 
underlying data and header information.

"""


default_app_config = "django_mri.apps.DjangoMriConfig"



from django_dicom.models import Series
from django_mri.models.nifti import NIfTI
from django_mri.models.scan import Scan
from django_mri.models.sequence_type import SequenceType
from django_mri.serializers.sequence_type import SequenceTypeSerializer
from django_mri.utils.utils import get_subject_model, get_group_model
from rest_framework import serializers


class ScanSerializer(serializers.HyperlinkedModelSerializer):
    url = serializers.HyperlinkedIdentityField(view_name="mri:scan-detail")
    dicom = serializers.HyperlinkedRelatedField(
        view_name="dicom:series-detail", queryset=Series.objects.all()
    )
    subject = serializers.HyperlinkedRelatedField(
        view_name="research:subject-detail",
        queryset=get_subject_model().objects.all(),
        required=False,
    )
    nifti = serializers.HyperlinkedRelatedField(
        view_name="mri:nifti-detail",
        queryset=NIfTI.objects.all(),
        required=False,
        allow_null=True,
    )
    study_groups = serializers.HyperlinkedRelatedField(
        view_name="research:group-detail",
        queryset=get_group_model().objects.all(),
        many=True,
        required=False,
    )
    sequence_type = SequenceTypeSerializer(
        source="infer_sequence_type_from_dicom", read_only=True
    )

    class Meta:
        model = Scan
        fields = (
            "id",
            "url",
            "dicom",
            "subject",
            "nifti",
            "study_groups",
            "institution_name",
            "time",
            "description",
            "number",
            "echo_time",
            "repetition_time",
            "inversion_time",
            "spatial_resolution",
            "comments",
            "sequence_type",
        )

    def create(self, validated_data):
        scan, created = Scan.objects.get_or_create(**validated_data)
        if created and scan.dicom and len(validated_data) == 1:
            scan.update_fields_from_dicom()
            scan.save()
        return scan



from django_filters import rest_framework as filters
from django_mri.models.sequence_type import SequenceType


class CharArrayFilter(filters.BaseCSVFilter, filters.CharFilter):
    pass


class SequenceTypeFilter(filters.FilterSet):
    scanning_sequence = CharArrayFilter("scanning_sequence", lookup_expr="exact")
    sequence_variant = CharArrayFilter("sequence_variant", lookup_expr="exact")

    class Meta:
        model = SequenceType
        fields = ("title", "scanning_sequence", "sequence_variant")



from rest_framework.pagination import PageNumberPagination


class StandardResultsSetPagination(PageNumberPagination):
    """
    Default pagination parameters. This didn't work as part of the DefaultsMixin and therefore has to be 
    defined separately in the 'pagination_class' configuration.

    """

    page_size = 25
    page_size_query_param = "page_size"



import os

from django.db.models.query import QuerySet
from django.conf import settings
from django.contrib.auth import get_user_model
from django.core.exceptions import ObjectDoesNotExist
from django.core.files.base import ContentFile
from django.core.files.storage import default_storage
from django_dicom.models import Series
from django_mri.data_import import ImportScan, LocalImport
from django_mri.filters.scan_filter import ScanFilter
from django_mri.models import Scan
from django_mri.serializers import ScanSerializer
from django_mri.utils import get_subject_model
from django_mri.views.defaults import DefaultsMixin
from django_mri.views.pagination import StandardResultsSetPagination
from rest_framework import status, viewsets
from rest_framework.decorators import action
from rest_framework.response import Response


class ScanViewSet(DefaultsMixin, viewsets.ModelViewSet):
    """
    API endpoint that allows scans to be viewed or edited.
    """

    pagination_class = StandardResultsSetPagination
    queryset = Scan.objects.all().order_by("-time")
    serializer_class = ScanSerializer
    filter_class = ScanFilter
    search_fields = (
        "id",
        "description",
        "number",
        "created",
        "scan_time",
        "echo_time",
        "inversion_time",
        "repetition_time",
        "sequence_type",
        "spatial_resolution",
        "institution_name",
        "is_updated_from_dicom",
    )
    ordering_fields = (
        "id",
        "description",
        "number",
        "created",
        "echo_time",
        "inversion_time",
        "repetition_time",
        "sequence_type",
        "spatial_resolution",
        "institution_name",
    )

    def get_queryset(self) -> QuerySet:
        """
        Filter the returned scans according to the studies the requesting
        user is a collaborator in, unless the user is staff, in which case
        return all scans.

        Returns
        -------
        QuerySet
            Scan instances.
        """

        user = get_user_model().objects.get(username=self.request.user)
        if user.is_staff:
            return Scan.objects.all()
        return Scan.objects.filter(study_groups__study__collaborators=user)

    @action(detail=False, methods=["POST"])
    def from_file(self, request):
        """
        Creates a new scan instance from a file (currently only DICOM format files are accepted).

        Parameters
        ----------
        request :
            A request from the client.

        Returns
        -------
        Response
            A response containing the serialized data or a message.
        """
        file_obj = request.data["file"]
        subject = get_subject_model().objects.get(id=request.data["subject_id"])
        user = get_user_model().objects.get(username=self.request.user)

        if file_obj.name.endswith(".dcm"):
            scan, created = ImportScan(
                subject, file_obj, scan_type="dcm", user=user
            ).run()
            serializer = ScanSerializer(scan, context={"request": request})
            return Response(serializer.data, status=status.HTTP_201_CREATED)

        elif file_obj.name.endswith(".zip"):
            content = ContentFile(file_obj.read())
            temp_file_name = default_storage.save("tmp.zip", content)
            temp_file_path = os.path.join(settings.MEDIA_ROOT, temp_file_name)
            LocalImport(subject, temp_file_path, user=user).run()
            os.remove(temp_file_path)
            return Response(
                {"message": "Successfully imported ZIP archive!"},
                status=status.HTTP_201_CREATED,
            )

    @action(detail=True, methods=["GET"])
    def from_dicom(self, request, series_id: int = None):
        """
        Returns scan information from a :class:`~django_dicom.models.series.Series` instance
        without serializing.

        Parameters
        ----------
        request :
            A request from the client.
        series_id : int, optional
            :class:`~django_dicom.models.series.Series` primary key, by default None

        Returns
        -------
        Response
            Serialized data or messagerequirements
        """
        try:
            series = Series.objects.get(id=series_id)
        except ObjectDoesNotExist:
            return Response(
                "Invalid DICOM primary key!", status=status.HTTP_400_BAD_REQUEST
            )
        try:
            scan = Scan.objects.get(dicom=series)
            return Response(
                ScanSerializer(scan, context={"request": request}).validated_data
            )
        except ObjectDoesNotExist:
            scan = Scan(dicom=series)
            scan.update_fields_from_dicom()
            serializer = ScanSerializer(scan, context={"request": request})
            return Response(serializer.data)



from django_filters.rest_framework import DjangoFilterBackend
from rest_framework.authentication import BasicAuthentication, TokenAuthentication
from rest_framework.filters import SearchFilter, OrderingFilter
from rest_framework.permissions import IsAuthenticated


class DefaultsMixin:
    """
    Default settings for view authentication, permissions and filtering.
    
    """

    authentication_classes = (BasicAuthentication, TokenAuthentication)
    permission_classes = (IsAuthenticated,)
    filter_backends = (DjangoFilterBackend, SearchFilter, OrderingFilter)



import os
from setuptools import find_packages, setup

with open("README.md", "r") as fh:
    long_description = fh.read()

with open("requirements.txt") as fh:
    requirements = fh.read().splitlines()

with open("requirements-dev.txt") as fh:
    dev_requirements = fh.read().splitlines()


# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name="django_mri",
    version="0.0.4",
    packages=find_packages(),
    include_package_data=True,
    license="AGPLv3",
    description="A reusable Django app to manage MRI data.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/TheLabbingProject/django_mri",
    author="Zvi Baratz",
    author_email="baratzz@pm.me",
    keywords="django mri neuroimaging research",
    python_requires=">=3.6",
    install_requires=requirements,
    extras_require={"dev": dev_requirements},
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Environment :: Web Environment",
        "Framework :: Django :: 2.2",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
)



from django import forms
from django.contrib.postgres.fields import ArrayField

# Copied from https://blogs.gnome.org/danni/2016/03/08/multiple-choice-using-djangos-postgres-arrayfield/


class ChoiceArrayField(ArrayField):
    """
    A field that allows us to store an array of choices.
     
    Uses Django 1.9's postgres ArrayField
    and a MultipleChoiceField for its formfield.
    """

    def formfield(self, **kwargs):
        defaults = {
            "form_class": forms.MultipleChoiceField,
            "choices": self.base_field.choices,
        }
        defaults.update(kwargs)
        # Skip our parent's formfield implementation completely as we don't
        # care for it.
        # pylint:disable=bad-super-call
        return super(ArrayField, self).formfield(**defaults)



from django.urls import path, include
from django_mri import views
from rest_framework import routers

app_name = "mri"
router = routers.DefaultRouter()
router.register(r"scan", views.ScanViewSet)
router.register(r"nifti", views.NiftiViewSet)
router.register(r"sequence_type", views.SequenceTypeViewSet)


urlpatterns = [
    path("mri/", include(router.urls)),
    path("mri/scan/from_file/", views.ScanViewSet.as_view({"POST": "from_file"})),
    path(
        "mri/scan/from_dicom/<int:series_id>/",
        views.ScanViewSet.as_view({"GET": "from_dicom"}),
    ),
    path("mri/scan/plot/<int:scan_id>/", views.ScanViewSet.as_view({"get": "plot"})),
]



from django.apps import apps
from django.conf import settings


def get_subject_model():
    app_label, model_name = settings.SUBJECT_MODEL.split(".")
    return apps.get_model(app_label=app_label, model_name=model_name)


def get_group_model():
    app_label, model_name = settings.STUDY_GROUP_MODEL.split(".")
    return apps.get_model(app_label=app_label, model_name=model_name)



from django_filters import rest_framework as filters
from django_mri.models.scan import Scan


class NumberInFilter(filters.BaseInFilter, filters.NumberFilter):
    pass


def filter_by_sequence_type(queryset, field_name, value):
    value = [int(pk) for pk in value]
    filtered_scan_ids = [
        scan.id
        for scan in queryset
        if scan.sequence_type and scan.sequence_type.id in value
    ]
    if -1 in value:
        filtered_scan_ids += [scan.id for scan in queryset if not scan.sequence_type]
    return queryset.filter(id__in=filtered_scan_ids)


class ScanFilter(filters.FilterSet):
    """
    Provides useful filtering options for the :class:`~django_dicom.models.series.Series`
    class.

    """

    description = filters.LookupChoiceFilter(
        "description",
        lookup_choices=[
            ("contains", "Contains (case-sensitive)"),
            ("icontains", "Contains (case-insensitive)"),
            ("exact", "Exact"),
        ],
    )
    number = filters.NumberFilter("number")
    scan_time = filters.DateTimeFromToRangeFilter("time")
    created = filters.DateTimeFromToRangeFilter("created")
    institution_name = filters.AllValuesFilter("institution_name")
    dicom_id_in = NumberInFilter(field_name="dicom__id", lookup_expr="in")
    sequence_type = NumberInFilter(method=filter_by_sequence_type)

    class Meta:
        model = Scan
        fields = (
            "id",
            "description",
            "number",
            "created",
            "scan_time",
            "echo_time",
            "inversion_time",
            "repetition_time",
            "institution_name",
            "is_updated_from_dicom",
            "dicom__id",
            "subject",
        )

